# copyright (C) 1997-98 Jean-Luc Fontaine (mailto:jfontain@mygale.org)
# this program is free software: please refer to the BSD type license enclosed in this package

set rcsId {$Id: blt2d.tcl,v 1.10 1998/04/19 13:54:27 jfontain Exp $}

class blt2DViewer {

    # use pie colors for consistency
    set blt2DViewer::(colors) {#7FFFFF #7FFF7F #FF7F7F #FFFF7F #7F7FFF #FFBF00 #BFBFBF #FF7FFF #FFFFFF}

    proc blt2DViewer {this path} viewer {} {
        $path configure -plotbackground $widget::(default,ButtonBackgroundColor)
        $path yaxis configure -title {} -min 0 -tickfont $font::(smallNormal)
        $path legend configure -borderwidth 1 -font $font::(mediumNormal)

        # allow dropping of data cells
        set blt2DViewer::($this,drop) [new dropSite\
            -path $path -formats DATACELLS -command "viewer::view $this \$dragSite::data(DATACELLS)"\
        ]

        set blt2DViewer::($this,drag) [new dragSite -path $path -validcommand "blt2DViewer::validateDrag $this"]
        dragSite::provide $blt2DViewer::($this,drag) OBJECTS "blt2DViewer::dragData $this"                ;# provide list of objects
        dragSite::provide $blt2DViewer::($this,drag) DATACELLS "blt2DViewer::dragData $this"

        set blt2DViewer::($this,elements) {}                                             ;# initialize list of data line identifiers
        set blt2DViewer::($this,selectedElements) {}
        bind $path <ButtonRelease-1> "blt2DViewer::select $this 0 %x %y"
        bind $path <Control-ButtonRelease-1> "blt2DViewer::select $this 1 %x %y"

        set blt2DViewer::($this,colorIndex) 0
        set blt2DViewer::($this,path) $path
    }

    proc ~blt2DViewer {this} {
        catch {trace vdelete $blt2DViewer::($this,array)(updates) w "blt2DViewer::update $this"}          ;# eventually remove trace
        delete $blt2DViewer::($this,drag) $blt2DViewer::($this,drop)
        eval delete $blt2DViewer::($this,elements)                                                       ;# delete existing elements
    }

    proc supportedTypes {this} {
        return {integer real}
    }

    proc dragData {this format} {
        switch $format {
            OBJECTS {
                if {[llength $blt2DViewer::($this,selectedElements)]>0} {
                    return $blt2DViewer::($this,selectedElements)                       ;# return selected elements if there are any
                } elseif {[llength $blt2DViewer::($this,elements)]==0} {
                    return $this                                                   ;# return graph itself if it contains no elements
                } else {
                    return {}                                                                            ;# return nothing otherwise
                }
            }
            DATACELLS {
                return [cellsFromElements $this $blt2DViewer::($this,selectedElements)]
            }
        }
    }

    proc validateDrag {this x y} {
        if {([llength $blt2DViewer::($this,selectedElements)]>0)||([llength $blt2DViewer::($this,elements)]==0)} {
            return 1                                                              ;# allow drag of selected elements or empty viewer
        } else {
            return 0                                                                                              ;# nothing to drag
        }
    }

    proc select {this control x y} {
        set element [$blt2DViewer::($this,path) legend get @$x,$y]
        if {[string length $element]==0} return                                                                   ;# not in a legend
        if {$control} {                                                                                  ;# toggle element selection
            set index [lsearch -exact $blt2DViewer::($this,selectedElements) $element]
            if {$index<0} {
                lappend blt2DViewer::($this,selectedElements) $element
                $blt2DViewer::($this,path) legend activate $element
            } else {
                set blt2DViewer::($this,selectedElements) [lreplace $blt2DViewer::($this,selectedElements) $index $index]
                $blt2DViewer::($this,path) legend deactivate $element
            }
        } else {                                                                                              ;# select sole element
            set blt2DViewer::($this,selectedElements) [list $element]
            $blt2DViewer::($this,path) legend deactivate *
            $blt2DViewer::($this,path) legend activate $element
        }
    }

    proc monitorCell {this array row column} {
        if {![info exists blt2DViewer::($this,array)]} {                                    ;# the module array should stay the same
            trace variable ${array}(updates) w "blt2DViewer::update $this"                    ;# refresh as module data gets updated
            set blt2DViewer::($this,array) $array
        }
        set cell ${array}($row,$column)
        if {[lsearch -exact [cellsFromElements $this $blt2DViewer::($this,elements)] $cell]>=0} return     ;# already charted, abort
        set element [newElement $this $blt2DViewer::($this,path) -label [viewer::label $array $row $column]]
        switched::configure $element -color [lindex $blt2DViewer::(colors) $blt2DViewer::($this,colorIndex)]
        switched::configure $element -deletecommand "blt2DViewer::deletedElement $this $element"  ;# keep track of element existence
        # circle through colors
        set blt2DViewer::($this,colorIndex) [expr {($blt2DViewer::($this,colorIndex)+1)%[llength $blt2DViewer::(colors)]}]
        lappend blt2DViewer::($this,elements) $element                                                     ;# register new data line
        set blt2DViewer::($this,cell,$element) $cell
    }

    proc deletedElement {this element} {
        ldelete blt2DViewer::($this,elements) $element
        catch {ldelete blt2DViewer::($this,selectedElements) $element}
        unset blt2DViewer::($this,cell,$element)
        if {[llength $blt2DViewer::($this,elements)]==0} {
            delete $this                                                            ;# self destruct when there are no more elements
        }
    }

    virtual proc update {this args}            ;# update display in derived class, using cells data. ignore eventual trace arguments

    virtual proc newElement {this path args}                                       ;# let derived class create an element of its own

    proc cellsFromElements {this elements} {
        set cells {}
        foreach element $elements {
            lappend cells $blt2DViewer::($this,cell,$element)
        }
        return $cells
    }
}
